from datetime import date, datetime

from django.http import HttpRequest
from django.utils.functional import cached_property
from wbcommission.models import Commission, Rebate


class CommissionPermissionMixin:
    queryset = Commission.objects.all()
    request: HttpRequest

    @cached_property
    def validity_date(self) -> date:
        if validity_date_repr := self.request.GET.get("validity_date"):
            return datetime.strptime(validity_date_repr, "%Y-%m-%d")
        return date.today()

    def get_queryset(self):
        # allow the user to see claim only on accounts it can see
        return (
            super().get_queryset().filter_for_user(self.request.user, validity_date=self.validity_date)  # type: ignore
        )


class RebatePermissionMixin:
    queryset = Rebate.objects.all()
    request: HttpRequest

    @cached_property
    def validity_date(self) -> date:
        if validity_date_repr := self.request.GET.get("validity_date"):
            return datetime.strptime(validity_date_repr, "%Y-%m-%d")
        return date.today()

    def get_queryset(self):
        # allow the user to see claim only on accounts it can see
        return (
            super().get_queryset().filter_for_user(self.request.user, validity_date=self.validity_date)  # type: ignore
        )



from django.dispatch import receiver
from rest_framework.reverse import reverse
from wbcore.contrib.directory.serializers import (
    CompanyModelSerializer,
    EntryModelSerializer,
    PersonModelSerializer,
)
from wbcore.filters.defaults import current_quarter_date_end, current_quarter_date_start
from wbcore.signals import add_instance_additional_resource


# register the rebate and generate report addition resources to the entry model serializers
@receiver(add_instance_additional_resource, sender=CompanyModelSerializer)
@receiver(add_instance_additional_resource, sender=PersonModelSerializer)
@receiver(add_instance_additional_resource, sender=EntryModelSerializer)
def commission_adding_additional_resource(sender, serializer, instance, request, user, **kwargs):
    start = current_quarter_date_start()
    end = current_quarter_date_end()
    res = {
        "rebates": f'{reverse("wbcommission:rebatetable-list", args=[], request=request)}?group_by=PRODUCT&recipient={instance.id}&date={start:%Y-%m-%d},{end:%Y-%m-%d}',
        "generate_customer_commission_report": f'{reverse("wbcommission:rebate-customerreport", request=request)}?recipient_id={instance.id}',
    }
    if user.has_perm("wbcommission.administrate_commission"):
        res[
            "generate_audit_commission_report"
        ] = f'{reverse("wbcommission:rebate-auditreport", request=request)}?recipient_id={instance.id}'
    return res



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class RebatePandasViewEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbcommission:rebatetable-list", request=self.request)


class RebateProductMarginalityEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_list_endpoint(self, **kwargs):
        return reverse("wbcommission:rebatemarginalitytable-list", request=self.request)

    def get_instance_endpoint(self, **kwargs):
        return reverse("wbportfolio:product-list", request=self.request)



from wbcommission.models import CommissionType
from wbcore import serializers as wb_serializers


class CommissionTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcommission:commissiontype-detail")

    class Meta:
        model = CommissionType
        fields = (
            "id",
            "name",
            "key",
            "_detail",
        )


class CommissionTypeModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = CommissionType
        read_only_fields = ("key",)
        fields = (
            "id",
            "name",
            "key",
        )



# Generated by Django 4.1.9 on 2023-06-28 12:21

from decimal import Decimal

import django.db.models.deletion
import wbcommission.models.rebate
from django.contrib.postgres.operations import BtreeGistExtension
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("wbcrm", "0005_account_accountrole_accountroletype_and_more"),
        ("wbportfolio", "0046_add_product_default_account"),
        ("directory", "0004_entry_is_draft_entry"),
    ]

    operations = [
        BtreeGistExtension(),
        migrations.CreateModel(
            name="AccountCustomer",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                (
                    "customer_type",
                    models.CharField(
                        choices=[("DIRECT", "Direct"), ("INDIRECT", "Indirect")],
                        default="DIRECT",
                        max_length=16,
                        verbose_name="Customer Type",
                    ),
                ),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="customers",
                        to="wbcrm.account",
                        verbose_name="Account",
                    ),
                ),
                (
                    "entry",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="account_customers",
                        to="directory.entry",
                        verbose_name="Customer",
                    ),
                ),
            ],
            options={
                "verbose_name": "Customer",
                "verbose_name_plural": "Customers",
                "notification_types": [
                    (
                        "wbcommission.accountcustomer.notify",
                        "New Account Customer Notification",
                        "Notifies you when a new account customer has been created",
                        True,
                        True,
                        False,
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="CommissionAccountRole",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                (
                    "role_type",
                    models.CharField(
                        choices=[("SALES", "Sales"), ("CUSTOMER_MANAGER", "Customer Manager")],
                        default="SALES",
                        max_length=16,
                        verbose_name="Role Type",
                    ),
                ),
                ("start", models.DateField(blank=True, null=True, verbose_name="Start")),
                ("end", models.DateField(blank=True, null=True, verbose_name="End")),
                ("weighting", models.FloatField(default=1, verbose_name="Weight")),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rebates_roles",
                        to="wbcrm.account",
                        verbose_name="Account Rebate roles",
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rebate_account_customer",
                        to="directory.entry",
                        verbose_name="Customer",
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="rebates_account_roles",
                        to="directory.person",
                        verbose_name="Person",
                    ),
                ),
            ],
            options={
                "verbose_name": "Commission Account Role",
                "verbose_name_plural": "Commission Account Roles",
            },
        ),
        migrations.CreateModel(
            name="AccountCustomerCommissionConstraint",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("start_year", models.PositiveIntegerField(blank=True, null=True)),
                ("end_year", models.PositiveIntegerField(blank=True, null=True)),
                ("payable_year_end", models.BooleanField(default=True)),
                ("payable_continous", models.BooleanField(default=True)),
                (
                    "customer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="constraints",
                        to="wbcommission.accountcustomer",
                    ),
                ),
                (
                    "products",
                    models.ManyToManyField(blank=True, related_name="customer_constraints", to="wbportfolio.product"),
                ),
            ],
            options={
                "verbose_name": "Account Customer Commission Constraint",
                "verbose_name_plural": "Account Customer Commission Constrains",
            },
        ),
        migrations.CreateModel(
            name="Rebate",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("date", models.DateField(verbose_name="Date")),
                (
                    "management_fees",
                    models.DecimalField(
                        decimal_places=4, default=Decimal(0.0), max_digits=16, verbose_name="Management Fees"
                    ),
                ),
                (
                    "pre_performance_fees",
                    models.DecimalField(
                        decimal_places=4,
                        default=Decimal(0.0),
                        max_digits=16,
                        verbose_name="Pre Performance Fees",
                    ),
                ),
                (
                    "performance_fees",
                    models.DecimalField(
                        decimal_places=4, default=Decimal(0.0), max_digits=16, verbose_name="Performance Fees"
                    ),
                ),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rebates",
                        to="wbcrm.account",
                        verbose_name="Account",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rebates",
                        to="wbportfolio.product",
                        verbose_name="Product",
                    ),
                ),
                (
                    "recipient",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="recipient_rebates",
                        to="directory.entry",
                        verbose_name="Recipient",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rebate",
                "verbose_name_plural": "Rebates",
                "unique_together": {("date", "recipient", "account", "product")},
                "index_together": {("date", "recipient", "product")},
            },
            bases=(wbcommission.models.rebate.BookingEntryCalculatedValueMixin, models.Model),
        ),
        migrations.CreateModel(
            name="Commission",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                (
                    "role_recipient",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("portfolio_managers", "Portfolio Managers"),
                            ("analysts", "Analysts"),
                            ("sales", "Sales"),
                            ("customer_managers", "Customer Managers"),
                        ],
                        max_length=32,
                        null=True,
                        verbose_name="Recipient Role",
                    ),
                ),
                ("order", models.IntegerField()),
                (
                    "commission_type",
                    models.CharField(
                        choices=[
                            ("net_management_fees", "Net Management Fees"),
                            ("gross_management_fees", "Gross Management Fees"),
                            ("net_performance_fees", "Net Performance Fees"),
                            ("gross_performance_fees", "Gross Performance Fees"),
                        ],
                        default="net_management_fees",
                        max_length=32,
                        verbose_name="Type",
                    ),
                ),
                ("start_date", models.DateField(verbose_name="Start Date")),
                (
                    "minimum_assets_under_management",
                    models.IntegerField(default=0, verbose_name="Minimum AUM"),
                ),
                ("percent", models.FloatField(verbose_name="Percent")),
                (
                    "account",
                    models.ForeignKey(
                        limit_choices_to=models.Q(("is_terminal_account", True)),
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="account_commissions",
                        to="wbcrm.account",
                        verbose_name="Account",
                    ),
                ),
                (
                    "crm_recipient",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="recipient_commissions",
                        to="directory.entry",
                        verbose_name="Recipient",
                    ),
                ),
            ],
            options={
                "verbose_name": "Commission",
                "verbose_name_plural": "Commissions",
                "unique_together": {
                    (
                        "account",
                        "order",
                        "crm_recipient",
                        "role_recipient",
                        "start_date",
                        "commission_type",
                        "minimum_assets_under_management",
                    )
                },
            },
        ),
    ]



from datetime import date
from decimal import Decimal
from io import BytesIO

import xlsxwriter
from celery import shared_task
from django.db.models import Case, F, Sum, When
from wbcommission.models import CommissionType, Rebate
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcrm.models.accounts import (
    Account,
    AccountRole,
    AccountRoleType,
    AccountRoleValidity,
)
from wbfdm.models import InstrumentPrice
from wbportfolio.models.transactions.claim import Claim
from xlsxwriter.utility import xl_rowcol_to_cell

from .utils import create_report_and_send

AccountCustomer = None


@shared_task(queue="commission")
def create_customer_report_and_send_as_task(user_id: int, recipient_id: int, start_date: date, end_date: date):
    user = User.objects.get(id=user_id)
    recipient = Entry.objects.get(id=recipient_id)
    create_report_and_send(user, recipient, start_date, end_date, create_report)


def create_report(user, customer, start_date, end_date):
    output = BytesIO()
    workbook = xlsxwriter.Workbook(output, {"in_memory": True})

    base_format = workbook.add_format({"font_name": "Liberation Sans", "font_size": 10})
    bold_format = workbook.add_format({"font_name": "Liberation Sans", "font_size": 10, "bold": True})
    decimal_format = workbook.add_format({"font_name": "Liberation Sans", "font_size": 10, "num_format": "#,##.00"})
    bold_decimal_format = workbook.add_format(
        {"font_name": "Liberation Sans", "font_size": 10, "num_format": "#,##.00", "bold": True}
    )
    percent_format = workbook.add_format({"font_name": "Liberation Sans", "font_size": 10, "num_format": "#,##.00 %"})
    bold_percent_format = workbook.add_format(
        {"font_name": "Liberation Sans", "font_size": 10, "num_format": "#,##.00 %", "bold": True}
    )

    # HERE STARTS THE FIRST WORKSHEET
    worksheet_products = workbook.add_worksheet(f"Products ({end_date:%d.%m.%Y})")

    worksheet_products.write_string(0, 0, "Product", bold_format)
    worksheet_products.write_string(0, 1, "Currency", bold_format)
    worksheet_products.write_string(0, 2, "Price", bold_format)
    worksheet_products.write_string(0, 3, "Price USD", bold_format)
    worksheet_products.write_string(0, 4, "Direct Assets", bold_format)
    worksheet_products.write_string(0, 5, "Direct Assets USD", bold_format)
    worksheet_products.write_string(0, 6, "Indirect Assets", bold_format)
    worksheet_products.write_string(0, 7, "Indirect Assets USD", bold_format)
    worksheet_products.write_string(0, 8, "Total Assets USD", bold_format)

    worksheet_products.write_formula(1, 5, "=SUM(F4:F99999)", bold_decimal_format)
    worksheet_products.write_formula(1, 7, "=SUM(H4:H99999)", bold_decimal_format)
    worksheet_products.write_formula(1, 8, "=SUM(I4:I99999)", bold_decimal_format)

    fx_rates = CurrencyFXRates.get_fx_rates_subquery(end_date)
    customer_accounts = Account.get_accounts_for_customer(customer)
    valid_customer_roles = (
        AccountRole.objects.filter(entry=customer)
        .annotate(is_currently_valid=AccountRoleValidity.get_role_validity_subquery(date.today()))
        .filter(is_currently_valid=True)
    )
    account_types = AccountRoleType.objects.filter(id__in=valid_customer_roles.values("role_type"))

    user_claims = Claim.objects.filter_for_customer(customer, include_related_roles=True).filter_for_user(
        user, validity_date=date.today()
    )
    products = (
        user_claims.filter(
            date__lte=end_date,
            status=Claim.Status.APPROVED,
        )
        .annotate(
            owner_asset_shares=Case(When(account__in=customer_accounts, then=F("shares")), default=Decimal(0)),
            **{
                f"{account_type.key}_assets_shares": Case(
                    When(
                        account__in=valid_customer_roles.filter(role_type=account_type).values("account"),
                        then=F("shares"),
                    ),
                    default=Decimal(0),
                )
                for account_type in account_types
            },
        )
        .values("product")
        .annotate(
            price=InstrumentPrice.subquery_closest_value(
                "net_value", val_date=end_date, instrument_pk_name="product__pk"
            ),
            product_id=F("product_id"),
            title=F("product__name"),
            currency=F("product__currency__symbol"),
            sum_shares=Sum("shares"),
            fx_rate=fx_rates,
            sum_owner_assets_shares=Sum("owner_asset_shares"),
            **{
                f"sum_{account_type.key}_assets_shares": Sum(f"{account_type.key}_assets_shares")
                for account_type in account_types
            },
        )
        .order_by("title")
    )
    for row, product in enumerate(products, start=3):
        product_cell = xl_rowcol_to_cell(row, 0)
        currency_cell = xl_rowcol_to_cell(row, 1)
        price_cell = xl_rowcol_to_cell(row, 2)
        price_usd_cell = xl_rowcol_to_cell(row, 3)
        owner_assets_cell = xl_rowcol_to_cell(row, 4)
        owner_assets_usd_cell = xl_rowcol_to_cell(row, 5)
        account_role_cells = dict()
        account_role_cells_usd = dict()
        index = 6
        for account_type in account_types:
            account_role_cells[account_type.key] = xl_rowcol_to_cell(row, index)
            account_role_cells_usd[account_type.key] = xl_rowcol_to_cell(row, index + 1)
            index += 2
        total_assets_usd_cell = xl_rowcol_to_cell(row, index)

        worksheet_products.write_string(product_cell, product["title"], base_format)
        worksheet_products.write_string(currency_cell, product["currency"], base_format)
        worksheet_products.write_number(price_cell, product["price"], decimal_format)
        worksheet_products.write_formula(price_usd_cell, f"{price_cell} * {product['fx_rate']}", decimal_format)

        worksheet_products.write_formula(
            owner_assets_cell, f"{product['sum_owner_assets_shares']} * {price_cell}", decimal_format
        )
        worksheet_products.write_formula(
            owner_assets_usd_cell, f"{owner_assets_cell} * {product['fx_rate']}", decimal_format
        )

        for account_type in account_types:
            worksheet_products.write_formula(
                account_role_cells[account_type.key],
                f"{product[f'sum_{account_type.key}_assets_shares']} * {price_cell}",
                decimal_format,
            )
            worksheet_products.write_formula(
                account_role_cells_usd[account_type.key],
                f"{account_role_cells[account_type.key]} * {product['fx_rate']}",
                decimal_format,
            )
        worksheet_products.write_formula(
            total_assets_usd_cell,
            " + ".join([owner_assets_usd_cell] + list(account_role_cells_usd.values())),
            decimal_format,
        )

    # HERE STARTS THE SECOND WORKSHEET
    worksheet_trade_performance = workbook.add_worksheet(f"Trades ({end_date:%d.%m.%Y})")

    worksheet_trade_performance.write_string(0, 0, "Trade Date", bold_format)
    worksheet_trade_performance.write_string(0, 1, "Shares", bold_format)
    worksheet_trade_performance.write_string(0, 2, "Bank", bold_format)
    worksheet_trade_performance.write_string(0, 3, "Root Account", bold_format)
    worksheet_trade_performance.write_string(0, 4, "Account", bold_format)
    worksheet_trade_performance.write_string(0, 5, "Product", bold_format)
    worksheet_trade_performance.write_string(0, 6, "Price (Trade Date)", bold_format)
    worksheet_trade_performance.write_string(0, 7, "Net Value (Trade Date)", bold_format)
    worksheet_trade_performance.write_string(0, 8, "Net Value (Trade Date) USD", bold_format)
    worksheet_trade_performance.write_string(0, 9, f"Price ({end_date:%d.%m.%Y})", bold_format)
    worksheet_trade_performance.write_string(0, 10, f"Net Value ({end_date:%d.%m.%Y})", bold_format)
    worksheet_trade_performance.write_string(0, 11, f"Net Value ({end_date:%d.%m.%Y}) USD", bold_format)
    worksheet_trade_performance.write_string(0, 12, "Performance", bold_format)

    worksheet_trade_performance.write_formula(1, 8, "=SUM(I4:I99999)", bold_decimal_format)
    worksheet_trade_performance.write_formula(1, 11, "=SUM(L4:L99999)", bold_decimal_format)
    worksheet_trade_performance.write_formula(1, 12, "=(L2/I2)-1", bold_percent_format)

    fx_rates_date = CurrencyFXRates.get_fx_rates_subquery("date")

    fx_rates_end = CurrencyFXRates.get_fx_rates_subquery(end_date)

    claims = (
        user_claims.filter(
            date__lte=end_date,
            status=Claim.Status.APPROVED,
        )
        .annotate(
            price_date=InstrumentPrice.subquery_closest_value("net_value", instrument_pk_name="product__pk"),
            price_end=InstrumentPrice.subquery_closest_value(
                "net_value", val_date=end_date, instrument_pk_name="product__pk"
            ),
            fx_rates_date=fx_rates_date,
            fx_rates_end=fx_rates_end,
        )
        .order_by("date")
    ).select_related("product", "account")

    for row, claim in enumerate(claims, start=3):
        trade_date_cell = xl_rowcol_to_cell(row, 0)
        shares_cell = xl_rowcol_to_cell(row, 1)
        bank_cell = xl_rowcol_to_cell(row, 2)
        root_account_cell = xl_rowcol_to_cell(row, 3)
        account_cell = xl_rowcol_to_cell(row, 4)
        product_cell = xl_rowcol_to_cell(row, 5)
        price_trade_cell = xl_rowcol_to_cell(row, 6)
        net_value_trade_cell = xl_rowcol_to_cell(row, 7)
        net_value_usd_trade_cell = xl_rowcol_to_cell(row, 8)
        price_end_cell = xl_rowcol_to_cell(row, 9)
        net_value_end_cell = xl_rowcol_to_cell(row, 10)
        net_value_usd_end_cell = xl_rowcol_to_cell(row, 11)
        performance_cell = xl_rowcol_to_cell(row, 12)

        worksheet_trade_performance.write_string(trade_date_cell, f"{claim.date:%d.%m.%Y}", base_format)
        worksheet_trade_performance.write_number(shares_cell, claim.shares, decimal_format)
        worksheet_trade_performance.write_string(bank_cell, claim.bank, base_format)
        worksheet_trade_performance.write_string(root_account_cell, claim.account.get_root().title, base_format)
        worksheet_trade_performance.write_string(account_cell, str(claim.account), base_format)
        worksheet_trade_performance.write_string(product_cell, claim.product.title, base_format)

        worksheet_trade_performance.write_number(
            price_trade_cell, claim.price_date or claim.product.share_price, decimal_format
        )
        worksheet_trade_performance.write_formula(
            net_value_trade_cell, f"={shares_cell}*{price_trade_cell}", decimal_format
        )
        worksheet_trade_performance.write_formula(
            net_value_usd_trade_cell, f"={net_value_trade_cell}*{claim.fx_rates_date}", decimal_format
        )

        worksheet_trade_performance.write_number(price_end_cell, claim.price_end, decimal_format)
        worksheet_trade_performance.write_formula(
            net_value_end_cell, f"={shares_cell}*{price_end_cell}", decimal_format
        )
        worksheet_trade_performance.write_formula(
            net_value_usd_end_cell, f"={net_value_end_cell}*{claim.fx_rates_end}", decimal_format
        )

        worksheet_trade_performance.write_formula(
            performance_cell, f"=({net_value_usd_end_cell}/{net_value_usd_trade_cell})-1", percent_format
        )

    # HERE STARTS THE THIRD WORKSHEET
    worksheet_rebates = workbook.add_worksheet(f"Rebates ({start_date:%d.%m.%Y}-{end_date:%d.%m.%Y})")

    worksheet_rebates.write_string(0, 0, "Date", bold_format)
    worksheet_rebates.write_string(0, 1, "Product", bold_format)
    worksheet_rebates.write_string(0, 2, "Account", bold_format)
    for index, commission_type in enumerate(CommissionType.objects.all()):
        worksheet_rebates.write_string(0, index + 3, f"{commission_type.name.title()} Fees", bold_format)
        worksheet_rebates.write_formula(
            1, index + 3, f"=SUM({chr(ord('@') + (index + 3))}3]:E99999)", bold_decimal_format
        )

    rebates_for_user = Rebate.objects.filter_for_user(user, validity_date=date.today())
    rebates = (
        rebates_for_user.filter(recipient=customer, date__gte=start_date, date__lte=end_date)
        .select_related("account", "product")
        .annotate(
            **{
                f"{commission_type.key}__value": Case(
                    When(commission_type=commission_type, then=F("value")), default=Decimal(0)
                )
                for commission_type in CommissionType.objects.all()
            }
        )
        .values("date", "account", "product")
        .annotate(
            product_title=F("product__name"),
            account_id=F("account__id"),
            account_title=F("account__computed_str"),
            **{
                f"sum_{commission_type.key}_value": Sum(f"{commission_type.key}__value")
                for commission_type in CommissionType.objects.all()
            },
        )
        .order_by("date", "product", "account_id")
    )

    for row, rebate in enumerate(rebates, start=4):
        date_cell = xl_rowcol_to_cell(row, 0)
        product_cell = xl_rowcol_to_cell(row, 1)
        account_cell = xl_rowcol_to_cell(row, 2)

        worksheet_rebates.write_string(date_cell, f"{rebate['date']:%d.%m.%Y}", base_format)
        worksheet_rebates.write_string(product_cell, rebate["product_title"], base_format)
        worksheet_rebates.write_string(account_cell, rebate["account_title"], base_format)
        for index, commission_type in enumerate(CommissionType.objects.all()):
            cell = xl_rowcol_to_cell(row, 3 + index)
            worksheet_rebates.write_number(cell, rebate[f"sum_{commission_type.key}_value"], decimal_format)

    workbook.close()
    output.seek(0)
    return (
        output,
        "customer_report_{}_{}_{}.xlsx".format(customer.computed_str, start_date, end_date),
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    )



import pytest
from django.db.models import Sum
from faker import Faker
from rest_framework.test import APIRequestFactory
from wbcommission.factories import (
    AccountTypeRoleCommissionFactory,
    CommissionFactory,
    RebateFactory,
)
from wbcommission.models.rebate import Rebate
from wbcommission.viewsets.rebate import RebateModelViewSet, RebatePandasView
from wbcore.contrib.authentication.factories.users import UserFactory
from wbcore.contrib.directory.models import Entry
from wbcore.utils.strings import format_number
from wbcrm.factories.accounts import AccountFactory, AccountRoleFactory

fake = Faker()


@pytest.mark.django_db
class TestRebateModelViewSet:
    @pytest.fixture
    def account_user(self):
        # True, we create a superuser

        # if fake.pybool():
        #     user = UserFactory.create(is_superuser=True)
        # else:
        user = UserFactory.create(is_superuser=False)
        entry = Entry.objects.get(id=user.profile.id)

        # Create a bunch of account and roles
        public_account = AccountFactory.create(is_public=True)
        role = AccountRoleFactory.create(account=public_account, entry=entry)
        RebateFactory.create(
            account=public_account,
            commission=AccountTypeRoleCommissionFactory.create(account_role_type_recipient=role.role_type),
        )  # shown rebate

        RebateFactory.create(commission=CommissionFactory.create(crm_recipient=entry))  # shown rebate
        RebateFactory.create()  # hidden rebate
        return user

    def test_ensure_permission_on_rebatetable(self, account_user):
        """
        We ensure that all claims viewset doesn't show more that what the user is allowed to see.
        For claim, the allowed claim are all the claims where the account is among the account they is allowed to see
        """
        allowed_rebates = Rebate.objects.filter_for_user(account_user)

        request = APIRequestFactory().get("")
        request.query_params = {}
        request.user = account_user
        viewset = RebatePandasView(request=request)
        assert allowed_rebates.exists()
        assert allowed_rebates.count() < Rebate.objects.count()  # Ensure that the filtering works
        assert set(allowed_rebates) == set(viewset.get_queryset())
        assert not viewset._get_dataframe().empty

    def test_ensure_permission_on_rebatemodelview(self, account_user):
        """
        We ensure that all claims viewset doesn't show more that what the user is allowed to see.
        For claim, the allowed claim are all the claims where the account is among the account they is allowed to see
        """
        allowed_rebates = Rebate.objects.filter_for_user(account_user)

        request = APIRequestFactory().get("")
        request.user = account_user
        viewset = RebateModelViewSet(request=request)
        queryset = viewset.get_queryset()
        assert allowed_rebates.exists()
        assert allowed_rebates.count() < Rebate.objects.count()  # Ensure that the filtering works
        assert set(allowed_rebates) == set(queryset)
        assert viewset.get_aggregates(queryset, queryset)["value_usd"]["Σ"] == float(
            format_number(allowed_rebates.aggregate(s=Sum("value"))["s"])
        )



from datetime import date as date_lib
from decimal import Decimal
from typing import Any, Optional

from celery import shared_task
from django.apps import apps
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Exists, OuterRef, QuerySet
from django.dispatch import receiver
from wbcore.contrib.authentication.models import User
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.signals import pre_merge
from wbcore.utils.enum import ChoiceEnum
from wbcrm.models.accounts import Account

from .commission import Commission, CommissionType


class BookingEntryCalculatedValueMixin:
    @classmethod
    def get_accounting_sum(cls, related_data: dict[str, str], queryset: QuerySet) -> float:
        raise NotImplementedError

    def update_calculated_value_of_booking_entry(self):
        raise NotImplementedError


class RebateGroupbyChoice(ChoiceEnum):
    ROOT_ACCOUNT = "Root Account"
    ACCOUNT = "Account"
    ROOT_ACCOUNT_OWNER = "Root Account Owner"
    ACCOUNT_OWNER = "Account Owner"
    PRODUCT = "Product"
    PRODUCT_GROUP = "ProductGroup"
    CLASSIFICATION = "Classification"
    RECIPIENT = "Recipient"

    @classmethod
    @property
    def map(cls) -> dict[str, dict[str, Any]]:
        """
        Field map used in the groupby filter in the rebate table view
        """
        return {
            "ROOT_ACCOUNT": {
                "pk": "root_account",
                "title_key": "root_account_repr",
                "search_fields": ["root_account_repr"],
            },
            "ACCOUNT": {
                "pk": "account",
                "title_key": "account__computed_str",
                "search_fields": ["account__computed_str"],
            },
            "ROOT_ACCOUNT_OWNER": {
                "pk": "root_account_owner",
                "title_key": "root_account_owner_repr",
                "search_fields": ["root_account_owner_repr"],
            },
            "ACCOUNT_OWNER": {
                "pk": "account__owner",
                "title_key": "account__owner__computed_str",
                "search_fields": ["account__owner__computed_str"],
            },
            "PRODUCT": {
                "pk": "product",
                "title_key": "product__computed_str",
                "search_fields": ["product__computed_str"],
            },
            "PRODUCT_GROUP": {
                "pk": "product__parent",
                "title_key": "product__parent__name",
                "search_fields": ["product__parent__name"],
            },
            "CLASSIFICATION": {
                "pk": "classification_id",
                "title_key": "classification_title",
                "search_fields": ["classification_title"],
            },
            "RECIPIENT": {
                "pk": "recipient__id",
                "title_key": "recipient__computed_str",
                "search_fields": ["recipient__computed_str"],
            },
        }


class RebateDefaultQueryset(QuerySet):
    def filter_for_user(self, user: User, validity_date: date_lib | None = None) -> QuerySet:
        """
        Protect the chained queryset and filter the rebates that this user cannot see based on the following rules:

        * not-hidden commission: Ever user with a direct valid account role on the commission's account
        * hidden: only user with a direct commission role on that commission line
        * in any case, all user with direct commission role
        """
        if not validity_date:
            validity_date = date_lib.today()
        if user.has_perm("wbcommission.administrate_commission"):
            return self
        allowed_commission_lines = Commission.objects.filter_for_user(user, validity_date=validity_date)
        return self.annotate(
            can_see_commission=Exists(allowed_commission_lines.filter(id=OuterRef("commission"))),
        ).filter(can_see_commission=True)


class RebateManager(models.Manager):
    def get_queryset(self) -> RebateDefaultQueryset:
        return RebateDefaultQueryset(self.model)

    def filter_for_user(self, user: User, validity_date: date_lib | None = None) -> QuerySet:
        return self.get_queryset().filter_for_user(user, validity_date=validity_date)


class Rebate(BookingEntryCalculatedValueMixin, models.Model):
    """The fees that get rebated to a recipient"""

    int: Optional[int]
    date = models.DateField(verbose_name="Date")
    account = models.ForeignKey(
        "wbcrm.Account",
        related_name="rebates",
        on_delete=models.CASCADE,
        verbose_name="Account",
        limit_choices_to=models.Q(("is_terminal_account", True)),
    )
    product = models.ForeignKey(
        "wbportfolio.Product", related_name="rebates", on_delete=models.CASCADE, verbose_name="Product"
    )
    recipient = models.ForeignKey(
        "directory.Entry", related_name="recipient_rebates", on_delete=models.PROTECT, verbose_name="Recipient"
    )
    commission_type = models.ForeignKey(
        "wbcommission.CommissionType",
        on_delete=models.PROTECT,
        related_name="rebates",
        verbose_name="Commission Type",
    )

    commission = models.ForeignKey(
        "wbcommission.Commission",
        related_name="rebates",
        on_delete=models.PROTECT,
        verbose_name="Commission Line",
    )

    value = models.DecimalField(max_digits=16, decimal_places=4, default=Decimal(0.0), verbose_name="Value")

    audit_log = models.JSONField(default=dict, verbose_name="Audit Log", encoder=DjangoJSONEncoder)
    objects = RebateManager()

    class Meta:
        verbose_name = "Rebate"
        verbose_name_plural = "Rebates"
        unique_together = ("date", "recipient", "account", "product", "commission_type")
        index_together = [
            # ("date", "recipient", "product"),
            ("commission_type", "date", "recipient", "product", "account"),
            #             ("date", "recipient", "product", "commission"),
        ]
        notification_types = [
            (
                "wbcommission.rebate.computation_done",
                "Rebate Computation Done",
                "Sends a notification to notify rebate computation requester that the calculation is done",
                True,
                True,
                False,
            ),
        ]

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"{self.date:%d.%m.%Y}: {self.recipient.computed_str} {self.account.title}"

    @classmethod
    def manage_rebate(
        cls,
        main_account: Account,
        prune_existing: bool = False,
        start_date: date_lib | None = None,
        only_content_object_ids: list[int] | None = None,
        **filter_kwargs,
    ):
        """
        Utility method to generate rebate for all commission types for the given root account.

        This method get the rebate information from the commission type's rebate manager and decide wether to create a new rebate or update an existing one

        Args:
            main_account: root account to generate rebates from
            prune_existing: If true, will delete all existing rebates for that account and its children. Default to False. Note: For optimization, the rebate manager will yield only valid rebate information and thus, existing obselete rebate might never be overidden with a zero value.
            **filter_kwargs: these key-word arguments are passed down the rebate manager for as keyword argument for the iterator function

        """
        if prune_existing:
            rebates_to_pruned = cls.objects.filter(account__in=main_account.get_descendants(include_self=True))
            if only_content_object_ids:
                rebates_to_pruned = rebates_to_pruned.filter(product__in=only_content_object_ids)
            if start_date:
                rebates_to_pruned = rebates_to_pruned.filter(date__gte=start_date)
            rebates_to_pruned.delete()
        updated_objs = []
        created_objs = []
        for commission_type in CommissionType.objects.all():
            for (
                terminal_account,
                compute_date,
                commission,
                content_object,
                recipient,
                recipient_fees,
                audit_log,
            ) in commission_type.compute_rebates(
                main_account, start_date=start_date, only_content_object_ids=only_content_object_ids, **filter_kwargs
            ):
                try:
                    rebate = Rebate.objects.get(
                        date=compute_date,
                        recipient=recipient,
                        account=terminal_account,
                        product=content_object,
                        commission_type=commission_type,
                    )
                    rebate.value = recipient_fees
                    rebate.commission = commission
                    rebate.audit_log = audit_log
                    updated_objs.append(rebate)
                except cls.DoesNotExist:
                    rebate = Rebate(
                        date=compute_date,
                        recipient=recipient,
                        account=terminal_account,
                        product=content_object,
                        commission_type=commission_type,
                        commission=commission,
                        value=recipient_fees,
                        audit_log=audit_log,
                    )
                    created_objs.append(rebate)

        cls.objects.bulk_update(updated_objs, ["value", "commission"], batch_size=10000)
        cls.objects.bulk_create(created_objs, batch_size=10000, ignore_conflicts=True)

    @classmethod
    def get_accounting_sum(cls, related_data: dict[str, str], queryset: QuerySet) -> float:
        return (queryset.aggregate(s=models.Sum(related_data["field"]))["s"] or 0) * -1

    def update_calculated_value_of_booking_entry(self):
        try:
            BookingEntry = apps.get_model("wbaccounting", "BookingEntry")  # type: ignore
            booking_entries = BookingEntry.objects.filter(  # type: ignore
                related_data__model="wbcommission.Rebate",
                related_data__data__date_gte__lte=self.date.strftime("%Y-%m-%d"),
                related_data__data__date_lte__gte=self.date.strftime("%Y-%m-%d"),
                related_data__data__recipient_id=self.recipient.id,
                related_data__data__product_id=self.product.id,
            )

            for booking_entry in booking_entries:
                booking_entry.calculated_value = booking_entry.get_related_data_accounting_sum()  # type: ignore
                booking_entry.save()  # type: ignore
        except LookupError:
            pass


# ----------- TASKS -----------


@shared_task(queue="commission")
def manage_rebate_as_task(
    main_account_id: int,
    start_date: date_lib | None = None,
    only_content_object_ids: list[int] | None = None,
    terminal_account_id: int | None = None,
    user: User | None = None,
    **kwargs,
):
    account = Account.objects.get(id=main_account_id)
    terminal_account_filter_dict = dict(id=terminal_account_id) if terminal_account_id else dict()
    Rebate.manage_rebate(
        main_account=account,
        start_date=start_date,
        only_content_object_ids=only_content_object_ids,
        terminal_account_filter_dict=terminal_account_filter_dict,
        **kwargs,
    )
    if user:
        send_notification(
            code="wbcommission.rebate.computation_done",
            title="Rebate Computation Done",
            body=f"The rebate computation for root account {account} is done",
            user=user,
        )


@receiver(pre_merge, sender="wbcrm.Account")
def handle_pre_merge_account_for_rebate(sender: models.Model, merged_object: Account, main_object: Account, **kwargs):
    """
    Aggregate the rebate if it exists already for the main account. Otherwise, reassign the account to point to the main account
    """
    for rebate in Rebate.objects.filter(account=merged_object).select_for_update():
        try:
            existing_rebate = Rebate.objects.get(
                date=rebate.date,
                recipient=rebate.recipient,
                account=main_object,
                product=rebate.product,
                commission_type=rebate.commission_type,
            )
            existing_rebate.value += rebate.value
            existing_rebate.save()
            rebate.delete()
        except Rebate.DoesNotExist:
            rebate.account = main_object
            rebate.save()



import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from rest_framework.reverse import reverse
from rest_framework.test import APIClient, APIRequestFactory
from wbcommission.analytics.marginality import MarginalityCalculator
from wbcommission.factories import CommissionTypeFactory, RebateFactory
from wbcommission.viewsets.rebate import RebateProductMarginalityViewSet
from wbportfolio.factories import FeesFactory, InstrumentPriceFactory, ProductFactory
from wbportfolio.models import Product

fake = Faker()


def _create_fixture(product, val_date, net_value=100, outstanding_shares=100):
    management = CommissionTypeFactory.create(key="management")
    perforance = CommissionTypeFactory.create(key="performance")
    i1 = InstrumentPriceFactory.create(
        instrument=product, net_value=net_value, outstanding_shares=outstanding_shares, calculated=False, date=val_date
    ).net_value  # create a price of AUM 100*100
    management_fees_1 = FeesFactory.create(
        linked_product=product, transaction_date=val_date, transaction_subtype="MANAGEMENT", calculated=False
    ).total_value
    performance_fees_1 = FeesFactory.create(
        linked_product=product, transaction_date=val_date, transaction_subtype="PERFORMANCE", calculated=False
    ).total_value
    performance_crys_fees_1 = FeesFactory.create(
        linked_product=product,
        transaction_date=val_date,
        transaction_subtype="PERFORMANCE_CRYSTALIZED",
        calculated=False,
    ).total_value
    management_rebate_1 = RebateFactory.create(product=product, date=val_date, commission_type=management).value
    performance_rebate_1 = RebateFactory.create(product=product, date=val_date, commission_type=perforance).value
    if val_date.weekday() == 0:
        return (
            i1,
            management_fees_1 / 3.0,
            performance_fees_1 / 3.0,
            performance_crys_fees_1 / 3.0,
            float(management_rebate_1) / 3.0,
            float(performance_rebate_1) / 3.0,
        )
    else:
        return (
            i1,
            management_fees_1,
            performance_fees_1,
            performance_crys_fees_1,
            float(management_rebate_1),
            float(performance_rebate_1),
        )


@pytest.mark.django_db
class TestMarginalityCalculator:
    @pytest.fixture
    def marginality_calculator(self):
        product = ProductFactory.create()
        start = (fake.date_object() + BDay(0)).date()
        end = (start + BDay(1)).date()

        (
            self.i1,
            self.management_fees_1,
            self.performance_fees_1,
            self.performance_crys_fees_1,
            self.management_rebate_1,
            self.performance_rebate_1,
        ) = _create_fixture(product, start)
        self.product_id = product.id
        self.start = start
        self.end = end
        return MarginalityCalculator(Product.objects.filter(id=product.id), start, end)

    def test_fees(self, marginality_calculator):
        assert marginality_calculator.management_fees.loc[self.product_id] == pytest.approx(
            self.management_fees_1, rel=1e-4
        )
        assert marginality_calculator.performance_fees.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1, rel=1e-4
        )
        assert marginality_calculator.total_fees.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1, rel=1e-4
        )

    def test_rebates(self, marginality_calculator):
        assert marginality_calculator.management_rebates.loc[self.product_id] == pytest.approx(
            self.management_rebate_1, rel=1e-4
        )
        assert marginality_calculator.performance_rebates.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1, rel=1e-4
        )
        assert marginality_calculator.total_rebates.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1 + self.management_rebate_1, rel=1e-4
        )

    def test_fees_usd(self, marginality_calculator):
        assert marginality_calculator.management_fees_usd.loc[self.product_id] == pytest.approx(
            self.management_fees_1, rel=1e-4
        )
        assert marginality_calculator.performance_fees_usd.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1, rel=1e-4
        )
        assert marginality_calculator.total_fees_usd.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1, rel=1e-4
        )

    def test_rebates_usd(self, marginality_calculator):
        assert marginality_calculator.management_rebates_usd.loc[self.product_id] == pytest.approx(
            self.management_rebate_1, rel=1e-4
        )
        assert marginality_calculator.performance_rebates_usd.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1, rel=1e-4
        )
        assert marginality_calculator.total_rebates_usd.loc[self.product_id] == pytest.approx(
            self.performance_rebate_1 + self.management_rebate_1, rel=1e-4
        )

    def test_marginality(self, marginality_calculator):
        assert marginality_calculator.management_marginality.loc[self.product_id] == pytest.approx(
            self.management_fees_1 - self.management_rebate_1, rel=1e-4
        )
        assert marginality_calculator.performance_marginality.loc[self.product_id] == pytest.approx(
            self.performance_fees_1 + self.performance_crys_fees_1 - self.performance_rebate_1, rel=1e-4
        )
        assert marginality_calculator.total_marginality.loc[self.product_id] == pytest.approx(
            (self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1)
            - (self.performance_rebate_1 + self.management_rebate_1),
            rel=1e-4,
        )

    def test_marginality_usd(self, marginality_calculator):
        assert marginality_calculator.management_marginality_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.management_marginality.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.performance_marginality_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.performance_marginality.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.total_marginality_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.total_marginality.loc[self.product_id], rel=1e-4
        )

    def test_marginality_percent(self, marginality_calculator):
        assert marginality_calculator.management_marginality_percent.loc[self.product_id] == pytest.approx(
            (self.management_fees_1 - float(self.management_rebate_1)) / self.management_fees_1, rel=1e-4
        )
        assert marginality_calculator.performance_marginality_percent.loc[self.product_id] == pytest.approx(
            (self.performance_fees_1 + self.performance_crys_fees_1 - float(self.performance_rebate_1))
            / (self.performance_fees_1 + self.performance_crys_fees_1),
            rel=1e-4,
        )
        assert marginality_calculator.total_marginality_percent.loc[self.product_id] == pytest.approx(
            (
                (self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1)
                - (float(self.performance_rebate_1) + float(self.management_rebate_1))
            )
            / (self.performance_fees_1 + self.performance_crys_fees_1 + self.management_fees_1),
            rel=1e-4,
        )

    def test_marginality_percent_usd(self, marginality_calculator):
        assert marginality_calculator.management_marginality_percent_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.management_marginality_percent.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.performance_marginality_percent_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.performance_marginality_percent.loc[self.product_id], rel=1e-4
        )
        assert marginality_calculator.total_marginality_percent_usd.loc[self.product_id] == pytest.approx(
            marginality_calculator.total_marginality_percent.loc[self.product_id], rel=1e-4
        )

    def test_get_net_marginality(self, marginality_calculator):
        assert marginality_calculator.get_net_marginality("management").loc[self.product_id] == pytest.approx(
            (self.management_fees_1 - float(self.management_rebate_1)) / (100 * 100) * 360, rel=1e-4
        )

    def test_get_aggregated_net_marginality(self, marginality_calculator):
        product_1 = ProductFactory.create()
        product_2 = ProductFactory.create()
        start = (fake.date_object() + BDay(0)).date()
        end = (start + BDay(1)).date()

        (
            i1,
            management_fees_1,
            performance_fees_1,
            performance_crys_fees_1,
            management_rebate_1,
            performance_rebate_1,
        ) = _create_fixture(product_1, start)
        (
            i2,
            management_fees_2,
            performance_fees_2,
            performance_crys_fees_2,
            management_rebate_2,
            performance_rebate_2,
        ) = _create_fixture(product_2, start, outstanding_shares=1000)
        total_aum = 100 * 100 + 100 * 1000
        calculator = MarginalityCalculator(Product.objects.filter(id__in=[product_1.id, product_2.id]), start, end)
        assert calculator.get_aggregated_net_marginality("management") == pytest.approx(
            ((management_fees_1 + management_fees_2) - (management_rebate_1 + management_rebate_2)) / total_aum * 360,
            rel=1e-4,
        )

    def test_rebate_marginality_view(self, marginality_calculator, super_user):
        request = APIRequestFactory().get("")
        request.query_params = {}
        request.GET = {"date_gte": self.start.strftime("%Y-%m-%d"), "date_lte": self.end.strftime("%Y-%m-%d")}
        request.user = super_user
        viewset = RebateProductMarginalityViewSet(request=request)
        assert set(viewset._get_dataframe().columns) == {
            "id",
            "title",
            "currency_symbol",
            "base_management_fees_percent",
            "management_fees",
            "management_rebates",
            "management_marginality",
            "management_marginality_percent",
            "base_performance_fees_percent",
            "performance_fees",
            "performance_rebates",
            "performance_marginality",
            "total_fees",
            "total_rebates",
            "total_marginality_percent",
            "total_fees_usd",
            "total_rebates_usd",
            "total_marginality_usd",
            "net_management_marginality",
            "net_performance_marginality",
        }

        # regression tests to check that without date range the view doesn't break
        url = reverse("wbcommission:rebatemarginalitytable-list", args=[])
        api_client = APIClient()
        api_client.force_authenticate(super_user)

        response = api_client.options(url)
        assert response.status_code == 200

        response = api_client.get(url)
        assert response.status_code == 200

        # check that with proper date range the viewset return some content
        response = api_client.options(url, data=request.GET)
        assert response.status_code == 200

        response = api_client.get(url, data=request.GET)
        assert response.status_code == 200
        assert response.json()["results"]



from unittest.mock import patch

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from faker import Faker
from wbcommission.models import CommissionType, Rebate

from .mixins import AccountManagerFixture

fake = Faker()


@pytest.mark.django_db
class TestRebateModel(AccountManagerFixture):
    def test_init(self, rebate):
        assert rebate.id is not None

    def test_rebate_view_permission(self, rebate_factory, user, commission_factory, commission_role_factory):
        # Here we simply test that if the user can see the underlying commission line, they can see the rebate (i.e. we check that the permission is applied for the queryset)
        # We do extensive unit testing on what commisison line a user can see on the corresponding test "test_filter_for_user"
        hidden_commission = commission_factory.create()
        hidden_rebate = rebate_factory.create(commission=hidden_commission)  # noqa

        commission_with_role = commission_factory.create()
        commission_role_factory.create(commission=commission_with_role, person=user.profile)
        visible_rebate = rebate_factory.create(commission=commission_with_role)
        assert set(Rebate.objects.filter_for_user(user)) == {visible_rebate}

        user.user_permissions.add(
            Permission.objects.get(content_type__app_label="wbcommission", codename="administrate_commission")
        )
        user = get_user_model().objects.get(id=user.id)
        assert set(Rebate.objects.filter_for_user(user)) == {visible_rebate, hidden_rebate}

    @patch.object(CommissionType, "compute_rebates")
    @pytest.mark.parametrize("rebate_value", [fake.pydecimal(right_digits=4, min_value=0, max_value=1000000)])
    def test_manage_rebate(self, mock_fct, commission_factory, fees_factory, rebate_value):
        fees = fees_factory.create()
        commission = commission_factory.create()
        # We check that the method create a rebate if none exists
        mock_fct.return_value = [
            (
                commission.account,
                fees.transaction_date,
                commission,
                fees.linked_product,
                commission.crm_recipient,
                rebate_value,
                dict(),
            )
        ]
        Rebate.manage_rebate(commission.account)
        new_rebate = Rebate.objects.get(
            commission=commission,
            product=fees.linked_product,
            date=fees.transaction_date,
            recipient=commission.crm_recipient,
        )
        assert new_rebate.value == rebate_value

        # we check that the method update the existing rebate
        mock_fct.return_value = [
            (
                commission.account,
                fees.transaction_date,
                commission,
                fees.linked_product,
                commission.crm_recipient,
                rebate_value * 2,
                dict(),
            )
        ]
        Rebate.manage_rebate(commission.account)
        new_rebate.refresh_from_db()
        assert new_rebate.value == rebate_value * 2

        # test that if rebate not valid anymore, a recomputation will automatically remove it
        new_commission = commission_factory.create(account=commission.account)
        mock_fct.return_value = [
            (
                commission.account,
                fees.transaction_date,
                new_commission,
                fees.linked_product,
                new_commission.crm_recipient,
                rebate_value,
                dict(),
            )
        ]
        Rebate.manage_rebate(commission.account, prune_existing=True)
        with pytest.raises((Rebate.DoesNotExist,)):
            new_rebate.refresh_from_db()
        assert (
            Rebate.objects.get(
                commission=new_commission,
                product=fees.linked_product,
                date=fees.transaction_date,
                recipient=new_commission.crm_recipient,
            ).value
            == rebate_value
        )

    def test_account_merging(self, account_factory, rebate_factory):
        """
        We want to test that:
        - Existing rebate for the same unique lookup arg are summed
        - If not existing rebate, then account are shifted
        """
        base_account = account_factory.create()
        merged_account = account_factory.create()
        base_rebate = rebate_factory.create(account=base_account)
        base_rebate.refresh_from_db()
        base_value = base_rebate.value  # for safekeeping

        # a rebate from the account that is going to be merged but with same date, recipient, product, and type. We expect this rebate to be deleted and its value summed to the base rebate
        merged_rebate_but_similar = rebate_factory.create(
            account=merged_account,
            date=base_rebate.date,
            recipient=base_rebate.recipient,
            product=base_rebate.product,
            commission_type=base_rebate.commission_type,
        )
        merged_rebate_but_similar.refresh_from_db()
        # rebate from the merged account but completly different: there won't be any existing rebate, hence the account will just be shifted.
        merged_rebate = rebate_factory.create(account=merged_account)

        base_account.merge(merged_account)
        base_rebate.refresh_from_db()
        assert base_rebate.value == base_value + merged_rebate_but_similar.value

        # check that the redundant rebate was deleted
        with pytest.raises(Rebate.DoesNotExist):
            merged_rebate_but_similar.refresh_from_db()
        merged_rebate.refresh_from_db()
        assert merged_rebate.account == base_account



from datetime import date
from decimal import Decimal

import numpy as np
import pandas as pd
from django.db.models import Case, OuterRef, Subquery, Value, When
from django.db.models.functions import Coalesce
from wbcommission.models import CommissionType, Rebate
from wbcore.contrib.currency.models import CurrencyFXRates
from wbfdm.models import InstrumentPrice
from wbportfolio.models import Fees


class MarginalityCalculator:
    FEE_MAP = {
        "MANAGEMENT": "management",
        "PERFORMANCE": "performance",
        "PERFORMANCE_CRYSTALIZED": "performance",
    }

    def __init__(self, products, from_date: date, to_date: date):
        products = products.annotate(
            fx_rate=Coalesce(
                Subquery(
                    CurrencyFXRates.objects.filter(
                        currency=OuterRef("currency"), date=OuterRef("last_valuation_date")
                    ).values("value")[:1]
                ),
                Decimal(1.0),
            )
        )

        self.fx_rates = (
            pd.DataFrame(products.values_list("id", "fx_rate"), columns=["id", "fx_rate"])
            .set_index("id")["fx_rate"]
            .astype(float)
        )

        # compute net marginality
        self.df_aum = pd.DataFrame(
            InstrumentPrice.objects.annotate_base_data()
            .filter(instrument__in=products, date__gte=from_date, date__lte=to_date)
            .values_list("calculated", "net_value_usd", "date", "outstanding_shares", "instrument"),
            columns=["calculated", "net_value_usd", "date", "outstanding_shares", "instrument"],
        ).rename(columns={"instrument": "id"})
        self.df_aum["date"] = pd.to_datetime(self.df_aum["date"])
        self.df_aum = (
            self.df_aum.sort_values(by="calculated")
            .groupby(["id", "date"])
            .agg({"net_value_usd": "last", "outstanding_shares": "first"})
        )
        self.df_aum = (self.df_aum.net_value_usd * self.df_aum.outstanding_shares).astype(float)
        self.df_aum = self.df_aum.reindex(
            pd.MultiIndex.from_product(
                [
                    self.df_aum.index.levels[0],
                    pd.date_range(
                        self.df_aum.index.get_level_values("date").min(),
                        self.df_aum.index.get_level_values("date").max(),
                    ),
                ],
                names=["id", "date"],
            ),
            method="ffill",
        ).dropna()

        # Build the fees dataframe where product id is the index and colum are the every fees type available and value are the amount.

        fees = Fees.valid_objects.filter(
            transaction_date__lte=to_date,
            transaction_date__gte=from_date,
            transaction_subtype__in=self.FEE_MAP.keys(),
            linked_product__in=products,
        ).annotate(
            fee_type=Case(
                *[When(transaction_subtype=k, then=Value(v)) for k, v in self.FEE_MAP.items()],
                default=Value("management"),
            )
        )
        self.df_fees = pd.DataFrame(
            fees.values_list("linked_product", "fee_type", "total_value", "transaction_date", "calculated"),
            columns=["linked_product", "fee_type", "total_value", "transaction_date", "calculated"],
        ).rename(columns={"linked_product": "id", "transaction_date": "date"})
        self.df_fees["date"] = pd.to_datetime(self.df_fees["date"])

        self.df_fees = (
            self.df_fees[["fee_type", "total_value", "id", "date"]]
            .pivot_table(index=["id", "date"], columns="fee_type", values="total_value", aggfunc="sum")
            .astype("float")
            .round(4)
        )
        self.df_fees["total"] = self.df_fees.sum(axis=1)
        self.df_fees = self.df_fees.reindex(self.df_aum.index, fill_value=0)
        self.df_fees = self._rolling_average_monday(self.df_fees)

        # Build the fees dataframe where product id is the index and colum are the every fees type available and value are the amount.
        self.df_rebates = pd.DataFrame(
            Rebate.objects.filter(date__gte=from_date, date__lte=to_date, product__in=products).values_list(
                "product", "value", "date", "commission_type__key"
            ),
            columns=["product", "value", "date", "commission_type__key"],
        ).rename(columns={"product": "id"})
        self.df_rebates["date"] = pd.to_datetime(self.df_rebates["date"])

        self.df_rebates = (
            pd.pivot_table(
                self.df_rebates,
                index=["id", "date"],
                columns="commission_type__key",
                values="value",
                aggfunc="sum",
                fill_value=0,
            )
            .astype("float")
            .round(4)
        )
        self.df_rebates = self.df_rebates.reindex(self.df_aum.index, fill_value=0)
        self.df_rebates["total"] = self.df_rebates.sum(axis=1)
        self.df_rebates = self._rolling_average_monday(self.df_rebates)
        # Iniliaze basic column

        self.empty_column = pd.Series(0.0, dtype="float64", index=self.df_fees.index)
        self._set_basics_statistics()

    def _set_basics_statistics(self):
        groupby_fees = self.df_fees.groupby(level=0).sum(numeric_only=True)
        groupby_rebates = self.df_rebates.groupby(level=0).sum(numeric_only=True)
        for key in [*CommissionType.objects.values_list("key", flat=True), "total"]:
            fees = groupby_fees.get(key, self.empty_column)
            rebates = groupby_rebates.get(key, self.empty_column)
            fees_usd = fees * self.fx_rates
            rebates_usd = rebates * self.fx_rates
            marginality = fees - rebates
            marginality_usd = fees_usd - rebates_usd
            marginality_percent = (fees - rebates) / fees.replace(0, np.nan)
            marginality_percent_usd = (fees_usd - rebates_usd) / fees_usd.replace(0, np.nan)

            setattr(self, f"{key}_fees", fees.rename(f"{key}_fees"))
            setattr(self, f"{key}_rebates", rebates.rename(f"{key}_rebates"))
            setattr(self, f"{key}_fees_usd", fees_usd.rename(f"{key}_fees_usd"))
            setattr(self, f"{key}_rebates_usd", rebates_usd.rename(f"{key}_rebates_usd"))
            setattr(self, f"{key}_marginality", marginality.rename(f"{key}_marginality"))
            setattr(self, f"{key}_marginality_usd", marginality_usd.rename(f"{key}_marginality_usd"))
            setattr(self, f"{key}_marginality_percent", marginality_percent.rename(f"{key}_marginality_percent"))
            setattr(
                self,
                f"{key}_marginality_percent_usd",
                marginality_percent_usd.rename(f"{key}_marginality_percent_usd"),
            )

    def _rolling_average_monday(self, df):
        """
        This utility method take a dataframe and assum the values on Mondays are accumulated over the weekend. So we need to average every Saturday, Sunday and Monday together.
        """

        monday = df[df.index.get_level_values("date").weekday == 0] / 3
        monday = monday.reindex(df.index, method="bfill")
        df[df.index.get_level_values("date").weekday.isin([5, 6, 0])] = monday[
            df.index.get_level_values("date").weekday.isin([5, 6, 0])
        ]
        return df

    def get_net_marginality(self, type: str) -> pd.Series:
        return (
            ((self.df_fees.get(type, self.empty_column) - self.df_rebates.get(type, self.empty_column)) / self.df_aum)
            .groupby("id")
            .mean()
        ) * 360

    def get_aggregated_net_marginality(self, type: str) -> pd.Series:
        # we compute the total net marginality for management for the aggregate function
        total_aum = self.df_aum.groupby(level=1).sum()
        return (
            (
                (
                    self.df_fees.get(type, self.empty_column).groupby(level=1).sum()
                    - self.df_rebates.get(type, self.empty_column).groupby(level=1).sum()
                )
                / total_aum
            ).mean(axis=0)
        ) * 360



import random
from decimal import Decimal

import pytest
from faker import Faker
from pandas.tseries.offsets import BDay
from wbcommission.models.account_service import AccountRebateManager
from wbportfolio.models import Claim

from .mixins import AccountManagerFixture

fake = Faker()


@pytest.mark.django_db
class TestAccountService(AccountManagerFixture):
    def test_get_commission_pool(
        self, management_account_manager, fees_factory, claim_factory, customer_trade_factory
    ):
        mngt_fees = fees_factory.create(transaction_subtype="MANAGEMENT")
        perf_fees = fees_factory.create(  # noqa
            transaction_subtype="PERFORMANCE",
            transaction_date=mngt_fees.transaction_date,
            linked_product=mngt_fees.linked_product,
        )  # noqa
        claim_factory.create(
            account=management_account_manager.root_account,
            trade=customer_trade_factory.create(underlying_instrument=mngt_fees.linked_product),
            status="APPROVED",
        )
        management_account_manager.initialize()

        assert (
            management_account_manager.get_commission_pool(mngt_fees.linked_product, mngt_fees.transaction_date)
            == mngt_fees.total_value
        )

    def test_get_commission_pool_performance(
        self, performance_account_manager, fees_factory, claim_factory, customer_trade_factory
    ):
        mngt_fees = fees_factory.create(transaction_subtype="MANAGEMENT")
        perf_fees = fees_factory.create(
            transaction_subtype="PERFORMANCE",
            transaction_date=mngt_fees.transaction_date,
            linked_product=mngt_fees.linked_product,
        )
        perf2_fees = fees_factory.create(
            transaction_subtype="PERFORMANCE_CRYSTALIZED",
            transaction_date=mngt_fees.transaction_date,
            linked_product=mngt_fees.linked_product,
        )

        claim_factory.create(
            account=performance_account_manager.root_account,
            trade=customer_trade_factory.create(underlying_instrument=mngt_fees.linked_product),
            status="APPROVED",
        )
        performance_account_manager.initialize()

        assert (
            performance_account_manager.get_commission_pool(mngt_fees.linked_product, mngt_fees.transaction_date)
            == perf_fees.total_value + perf2_fees.total_value
        )

    def test_get_commission_pool_with_calculated(
        self, management_account_manager, fees_factory, claim_factory, customer_trade_factory
    ):
        # check that in case there is only calculate, we return the calculated fees value, but in case there is calculated and real fees, we use the real fees value
        calculated_fees = fees_factory.create(calculated=True, transaction_subtype="MANAGEMENT")
        claim_factory.create(
            account=management_account_manager.root_account,
            trade=customer_trade_factory.create(underlying_instrument=calculated_fees.linked_product),
            status="APPROVED",
        )
        management_account_manager.initialize()

        assert (
            management_account_manager.get_commission_pool(
                calculated_fees.linked_product, calculated_fees.transaction_date
            )
            == calculated_fees.total_value
        )
        # Check that is there is a non calculated fees, it is used instead of the calculated one
        fees = fees_factory.create(transaction_subtype=calculated_fees.transaction_subtype, calculated=False)
        claim_factory.create(
            account=management_account_manager.root_account,
            trade=customer_trade_factory.create(underlying_instrument=fees.linked_product),
            status="APPROVED",
        )
        management_account_manager.initialize()

        assert (
            management_account_manager.get_commission_pool(fees.linked_product, fees.transaction_date)
            == fees.total_value
        )

    @pytest.mark.parametrize("val_date", [fake.past_date()])
    def test_get_terminal_account_holding_ratio(
        self, account_factory, claim_factory, instrument_price_factory, product_factory, val_date, commission_type
    ):
        product1 = product_factory.create()
        product2 = product_factory.create()
        val_date = (val_date + BDay(0)).date()
        next_val_date = (val_date + BDay(1)).date()

        product1_price_val_date = instrument_price_factory.create(
            date=val_date, instrument=product1, outstanding_shares=Decimal(2e4), calculated=True
        )

        product1_price_next_val_date = instrument_price_factory.create(
            date=next_val_date, instrument=product1, outstanding_shares=Decimal(2e4), calculated=True
        )

        product2_price_val_date = instrument_price_factory.create(
            date=val_date, instrument=product2, outstanding_shares=Decimal(2e4), calculated=True
        )
        product2_price_next_val_date = instrument_price_factory.create(
            date=next_val_date, instrument=product2, outstanding_shares=Decimal(2e4), calculated=True
        )

        parent_account = account_factory.create(is_terminal_account=False, is_active=True)
        claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=parent_account,
        )  # ignore this claim because it's attached to a non terminal account

        child_terminal_account = account_factory.create(
            parent=parent_account, is_terminal_account=True, is_active=True
        )
        claim_factory.create(  # ignore this claim because it's not approved
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            account=child_terminal_account,
            trade__underlying_instrument=product1,
            status=random.choice([Claim.Status.DRAFT, Claim.Status.WITHDRAWN, Claim.Status.PENDING]),
        )
        unvalid_child_terminal_account = account_factory.create(parent=parent_account, is_active=False)
        claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=unvalid_child_terminal_account,
        )  # ignore this claim as the attached account is not active

        valid_claim1 = claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=child_terminal_account,
        )
        valid_claim2 = claim_factory.create(
            trade__value_date=next_val_date,
            trade__transaction_date=(next_val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=child_terminal_account,
        )
        valid_claim3 = claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product2,
            status=Claim.Status.APPROVED,
            account=child_terminal_account,
        )
        product1_price_val_date.refresh_from_db()
        product1_price_next_val_date.refresh_from_db()
        product2_price_val_date.refresh_from_db()
        product2_price_next_val_date.refresh_from_db()

        account_manager = AccountRebateManager(parent_account, commission_type.key)
        account_manager.initialize()
        assert account_manager.get_terminal_account_holding_ratio(
            child_terminal_account, product1, val_date
        ) == pytest.approx((valid_claim1.shares / product1_price_val_date.outstanding_shares), rel=Decimal(1e-6))
        assert account_manager.get_terminal_account_holding_ratio(
            child_terminal_account, product1, next_val_date
        ) == pytest.approx(
            (valid_claim1.shares + valid_claim2.shares) / product1_price_next_val_date.outstanding_shares,
            rel=Decimal(1e-6),
        )
        assert account_manager.get_terminal_account_holding_ratio(
            child_terminal_account, product2, val_date
        ) == pytest.approx((valid_claim3.shares / product2_price_val_date.outstanding_shares), rel=Decimal(1e-6))
        assert account_manager.get_terminal_account_holding_ratio(
            child_terminal_account, product2, next_val_date
        ) == pytest.approx((valid_claim3.shares / product2_price_next_val_date.outstanding_shares), rel=Decimal(1e-6))

    @pytest.mark.parametrize("val_date", [fake.date_object()])
    def test_get_root_account_total_holding(
        self, account_factory, claim_factory, product_factory, instrument_price_factory, val_date, commission_type
    ):
        product1 = product_factory.create()
        product2 = product_factory.create()
        val_date = (val_date + BDay(0)).date()
        next_val_date = (val_date + BDay(1)).date()
        product1_price_val_date = instrument_price_factory.create(date=val_date, instrument=product1, calculated=False)
        product1_price_next_val_date = instrument_price_factory.create(
            date=next_val_date, instrument=product1, calculated=False
        )
        product2_price_val_date = instrument_price_factory.create(date=val_date, instrument=product2, calculated=False)
        product2_price_next_val_date = instrument_price_factory.create(
            date=next_val_date, instrument=product2, calculated=False
        )

        parent_account = account_factory.create(is_terminal_account=False, is_active=True)
        claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=parent_account,
        )  # ignore this claim because it's attached to a non terminal account

        child_terminal_account = account_factory.create(
            parent=parent_account, is_terminal_account=True, is_active=True
        )
        claim_factory.create(  # ignore this claim because it's not approved
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            account=child_terminal_account,
            trade__underlying_instrument=product1,
            status=random.choice([Claim.Status.DRAFT, Claim.Status.WITHDRAWN, Claim.Status.PENDING]),
        )
        unvalid_child_terminal_account = account_factory.create(parent=parent_account, is_active=False)
        claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=unvalid_child_terminal_account,
        )  # ignore this claim as the attached account is not active

        valid_claim1 = claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=child_terminal_account,
        )
        valid_claim2 = claim_factory.create(
            trade__value_date=next_val_date,
            trade__transaction_date=(next_val_date - BDay(1)).date(),
            trade__underlying_instrument=product1,
            status=Claim.Status.APPROVED,
            account=child_terminal_account,
        )
        valid_claim3 = claim_factory.create(
            trade__value_date=val_date,
            trade__transaction_date=(val_date - BDay(1)).date(),
            trade__underlying_instrument=product2,
            status=Claim.Status.APPROVED,
            account=child_terminal_account,
        )

        account_manager = AccountRebateManager(parent_account, commission_type.key)
        account_manager.initialize()

        valid_claim1_aum = (  # noqa
            valid_claim1.shares
            * product1_price_val_date.net_value
            * product1_price_val_date.currency_fx_rate_to_usd.value
        )
        valid_claim2_aum = (  # noqa
            valid_claim2.shares
            * product1_price_next_val_date.net_value
            * product1_price_next_val_date.currency_fx_rate_to_usd.value
        )
        valid_claim3_aum = (  # noqa
            valid_claim3.shares
            * product2_price_val_date.net_value
            * product2_price_val_date.currency_fx_rate_to_usd.value
        )
        assert account_manager.get_root_account_total_holding(val_date) == (
            pytest.approx(
                (
                    valid_claim1.shares * product1_price_val_date._net_value_usd
                    + valid_claim3.shares * product2_price_val_date._net_value_usd
                ),
                rel=Decimal(1e-6),
            )
        )
        assert account_manager.get_root_account_total_holding(next_val_date) == (
            pytest.approx(
                (valid_claim1.shares + valid_claim2.shares) * product1_price_next_val_date._net_value_usd
                + valid_claim3.shares * product2_price_next_val_date._net_value_usd,
                rel=Decimal(1e-8),
            )
        )
